# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
from dash import dcc, html, Input, Output, State

import odatix.gui.ui_components as ui
import odatix.gui.navigation as navigation
from odatix.lib.settings import OdatixSettings

page_path = "/"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix',
    name='Home',
)

######################################
# UI Components
######################################

padding = 20

# Home page

home_cards = [
    {
        "name": "Architectures",
        "link": "/architectures",
        "image": "assets/icons/architecture.png",
        "description": "Configure your architectures",
    },
    # {
    #     "name": "Run jobs",
    #     "link": "/run",
    #     "image": "assets/icons/run.png",
    #     "description": "Run synthesis and simulation",
    # },
    # {
    #     "name": "Export Results",
    #     "link": "/export",
    #     "image": "assets/icons/export.png",
    #     "description": "Export results from synthesis and simulation",
    # },
    # {
    #     "name": "Explore Results",
    #     "link": "/explorer",
    #     "image": "assets/icons/explorer.png",
    #     "description": "Explore results in an interactive interface",
    # },
    # {
    #     "name": "EDA Tools",
    #     "link": "/tools",
    #     "image": "assets/icons/tools.png",
    #     "description": "EDA tools options and metrics definition",
    # },
    {
        "name": "Workspace Settings",
        "link": "/workspace",
        "image": "assets/icons/settings.png",
        "description": "Adjust workspace settings",
    },
    {
        "name": "Documentation",
        "link": "https://odatix.readthedocs.io",
        "image": "assets/icons/documentation.png",
        "description": "Access Odatix online documentation",
    },
]

home_layout = [
    html.Div(
        children=[
            html.Div(
                [ui.create_card_button(card) for card in home_cards],
                style={
                    "display": "flex",
                    "flexWrap": "wrap",
                    "justifyContent": "center",
                    "gap": "20px",
                    "paddingLeft": "50px",
                    "paddingRight": "50px",
                },
            ),
        ],
        id=f"{__name__}-content",
        style={
            "display": "flex",
            "justifyContent": "center",
            "alignItems": "center",
            "paddingTop": f"{padding}px",
            "paddingBottom": f"{padding}px",
        },
    ),
    html.H4(
        "Icons designed by Freepik from Flaticon",
        style={
            "textAlign": "center",
            "color": "#aaa",
            "marginTop": "20px",
            "fontSize": "12px",
            "fontWeight": "normal",
        }
    ),
]


# New workspace page

new_workspace_cards = [
    {
        "name": "Empty workspace",
        "id": "create-empty-workspace",
        "image": "assets/icons/workspace_empty.svg",
        "description": "Start from scratch",
    },
    {
        "name": "Workspace with examples",
        "id": "create-workspace-with-examples",
        "image": "assets/icons/workspace_examples.svg",
        "description": "Get started quickly with pre-configured examples",
    },
]

new_workspace_layout = [
    html.Div(
        children=[
            html.H2("Initialize an Odatix workspace in this directory", style={"fontSize": "32px", "marginBottom": "-15px"}),
            html.Pre(
                f"{os.getcwd()}", 
                style={
                    "display": "inline-block",
                    "width": "fit-content",
                    "fontSize": "20px",
                    "padding": "0px 10px",
                    "whiteSpace": "pre-wrap",
                    "overflowWrap": "anywhere",
                }
            ),
        ],
        className="tile center",
        style={"fontSize": "32px"}
    ),
    html.Div(
        [
            html.Div(
                [html.Div(style={"display": "none"})] + [ui.create_card_button(card) for card in new_workspace_cards],
                style={
                    "display": "flex",
                    "flexWrap": "wrap",
                    "justifyContent": "center",
                    "gap": "20px",
                    "paddingLeft": "50px",
                    "paddingRight": "50px",
                },
            ),
        ],
        id=f"{__name__}-content",
        style={
            "display": "flex",
            "justifyContent": "center",
            "alignItems": "center",
            "paddingTop": f"{padding}px",
            "paddingBottom": f"{padding}px",
        },
    ),
    html.H4(
        "Icons designed by Freepik from Flaticon",
        className="subtle-text",
        style={
            "textAlign": "center",
            "marginTop": "20px",
        }
    ),
    html.Div(style={"height": "10vh"}),  # Spacer at the bottom
]


######################################
# Callbacks
######################################

@dash.callback(
    Output("home-page-content", "children"),
    Input("url", "pathname"),
    Input("odatix-settings", "data"),
    prevent_initial_call=False,
)
def redirect_to_new_workspace(search, odatix_settings):
    triggered_id = dash.ctx.triggered_id
    if odatix_settings == {}:
        return new_workspace_layout
    else:
        return home_layout

@dash.callback(
    Output("odatix-settings", "data", allow_duplicate=True),
    Input("create-empty-workspace", "n_clicks"),
    Input("create-workspace-with-examples", "n_clicks"),
    prevent_initial_call=True,
)
def create_empty_workspace(n_clicks_empty, n_clicks_examples):
    triggered_id = dash.ctx.triggered_id
    success = False
    if triggered_id == "create-empty-workspace" and n_clicks_empty:
        success = OdatixSettings.init_directory_nodialog(include_examples=False, silent=True),
    elif triggered_id == "create-workspace-with-examples" and n_clicks_examples:
        success = OdatixSettings.init_directory_nodialog(include_examples=True, silent=True),
    if success: 
        odatix_settings = OdatixSettings()
        return odatix_settings.to_dict()
    return dash.no_update

######################################
# Layout
######################################

layout = html.Div(
    children=[
        dcc.Location(id="url", refresh=False),
        html.Div(id="home-page-content"),
    ],
    id="home-page",
    className="page-content",
    style={
        "width": "100%", 
        "minHeight": f"calc(100vh - {navigation.top_bar_height})",
        "display": "flex",  
        "flexDirection": "column",
        "justifyContent": "center",
        "alignItems": "center",
    }
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import curses
import re

A_NORMAL = 0


class AnsiToCursesConverter:
  LIGHT_OFFSET = 8

  # Mapping of ANSI color codes to curses color pairs
  ANSI_COLORS = {
    "30": curses.COLOR_BLACK,
    "31": curses.COLOR_RED,
    "32": curses.COLOR_GREEN,
    "33": curses.COLOR_YELLOW,
    "34": curses.COLOR_BLUE,
    "35": curses.COLOR_MAGENTA,
    "36": curses.COLOR_CYAN,
    "37": curses.COLOR_WHITE,
    "90": curses.COLOR_BLACK + LIGHT_OFFSET,  # Light black (grey)
    "91": curses.COLOR_RED + LIGHT_OFFSET,  # Light red
    "92": curses.COLOR_GREEN + LIGHT_OFFSET,  # Light green
    "93": curses.COLOR_YELLOW + LIGHT_OFFSET,  # Light yellow
    "94": curses.COLOR_BLUE + LIGHT_OFFSET,  # Light blue
    "95": curses.COLOR_MAGENTA + LIGHT_OFFSET,  # Light magenta
    "96": curses.COLOR_CYAN + LIGHT_OFFSET,  # Light cyan
    "97": curses.COLOR_WHITE + LIGHT_OFFSET,  # Light white
  }

  def __init__(self):
    self.current_color = None
    self.current_intensity = 0
    self.initialized = False

  def initialize_colors(self):
    """
    Initialize color pairs in curses.
    """
    if not self.initialized:
      curses.start_color()
      curses.use_default_colors()
      for code, color in AnsiToCursesConverter.ANSI_COLORS.items():
        if color != -1:
          curses.init_pair(int(code), color, -1)
      self.current_color = curses.color_pair(0)
      self.initialized = True

  def reset_format(self):
    """
    Reset the current formatting attributes.

    This method resets both the color and intensity attributes to their default values,
    ensuring that subsequent text is displayed without inherited styles.

    Effects:
        - Sets `self.current_color` to the default color pair (curses.color_pair(0)).
        - Sets `self.current_intensity` to `A_NORMAL`.

    Use this method when:
        - Resetting styles after processing ANSI escape sequences.
        - Ensuring consistent formatting when switching contexts.
    """
    self.current_color = curses.color_pair(0)
    self.current_intensity = A_NORMAL

  def add_ansi_str(self, win, text, debug_win=None, width=-1, dim=False):
    """
    Add a string with ANSI escape sequences to a curses window.

    Args:
        win (curses.window): The window where text is displayed.
        text (str): The text containing ANSI escape sequences.
        debug_win (curses.window, optional): Window for debugging output.
        width (int, optional): Maximum width of the displayed text.
        dim (bool, optional): Apply dim effect.
    """
    self.initialize_colors()

    # Regex to match ANSI escape sequences
    ansi_escape = re.compile(r"\x1b\[([0-9;]*)m")

    # Split text by ANSI codes
    segments = ansi_escape.split(text)

    current_width = 0
    truncated = False
    dim = dim and curses.A_DIM

    # Clear line before writing new text to prevent color bleeding
    win.clrtoeol()

    for i, segment in enumerate(segments):
      if i % 2 == 0:
        # Normal text
        if width > 0:
          remaining_width = width - current_width
          if remaining_width > 0:
            if len(segment) > remaining_width:
              segment = segment[:remaining_width-3]
              truncated = True
            current_width += len(segment)
            win.addstr(segment, self.current_color | self.current_intensity | dim)
          if truncated:
            win.addstr("...", curses.color_pair(90) | dim)
        else:
          win.addstr(segment, self.current_color | self.current_intensity | dim)
      else:
        # ANSI code processing
        codes = segment.split(";")
        for code in codes:
          if code in AnsiToCursesConverter.ANSI_COLORS:
            color_pair = curses.color_pair(int(code))
            self.current_color = color_pair
          elif code == "0":  # Reset all attributes
            self.current_color = curses.color_pair(0)
            self.current_intensity = A_NORMAL
          elif code == "1":  # Bold
            self.current_intensity = curses.A_BOLD
          elif code == "22":  # Normal intensity
            self.current_intensity = A_NORMAL
          else:
            if debug_win is not None:
              debug_win.addstr(f"Unsupported ANSI code: {code}", curses.color_pair(31) | dim)

    # Refresh the window after updates
    win.refresh()
    if debug_win is not None:
      debug_win.refresh()

  def test(self, stdscr):
    """
    Test function to display formatted text with ANSI escape sequences.

    Args:
        stdscr (curses.window): The standard screen window.
    """
    stdscr.clear()
    text = " default_text\n \x1b[1m\x1b[31mbold_red_text\x1b[0m\n \x1b[34mblue_and_\x1b[32mgreen_text\n \x1b[31mred_text\n \x1b[0mdefault_text\n \x1b[52merror_text"
    height, width = stdscr.getmaxyx()

    # Ensure the text fits within the window width
    self.add_ansi_str(stdscr, text, width=width)

    # Wait for user input
    stdscr.getch()


if __name__ == "__main__":
  conv = AnsiToCursesConverter()
  curses.wrapper(conv.test)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import shutil
import yaml
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap, CommentedSeq
from datetime import datetime
from itertools import product

import odatix.components.motd as motd
import odatix.lib.hard_settings as hard_settings
from odatix.lib.utils import copytree
from typing import Optional


######################################
# Architectures and Simulations
######################################

def get_instances(path: str) -> list:
    """
    Get the list of architectures or simulations.
    """
    if not os.path.exists(path):
        return []
    return sorted([
        d for d in os.listdir(path)
        if os.path.isdir(os.path.join(path, d))
    ])

def get_simulations(path: str) -> list:
    """
    Get the list of simulations.
    """
    return get_instances(path)

def get_architectures(path: str) -> list:
    """
    Get the list of architectures.
    """
    return get_instances(path)

def instance_exists(path, name) -> bool:
    """
    Check if a specific architecture or simulation exists.
    """
    path = os.path.join(path, name)
    return os.path.isdir(path)

def architecture_exists(path, name) -> bool:
    """
    Check if a specific architecture exists.
    """
    return instance_exists(path, name)

def simulation_exists(path, name) -> bool:
    """
    Check if a specific simulation exists.
    """
    return instance_exists(path, name)

def duplicate_instance(path, source_name, target_name) -> None:
    """
    Duplicate an architecture or simulation.
    """
    source_path = os.path.join(path, source_name)
    target_path = os.path.join(path, target_name)
    if not os.path.isdir(source_path):
        raise ValueError(f"Source instance '{source_name}' does not exist.")
    if os.path.exists(target_path):
        raise ValueError(f"Target instance '{target_name}' already exists.")
    copytree(source_path, target_path)

def delete_instance(path, name) -> None:
    """
    Delete an architecture or simulation.
    """
    path = os.path.join(path, name)
    if os.path.isdir(path):
        shutil.rmtree(path)
    else:
        os.remove(path)

def rename_instance(path, old_name, new_name) -> None:
    """
    Rename an architecture or simulation.
    """
    old_path = os.path.join(path, old_name)
    new_path = os.path.join(path, new_name)
    if not os.path.isdir(old_path):
        return
    if os.path.exists(new_path):
        return
    shutil.move(old_path, new_path)

def rename_architecture(path, old_name, new_name) -> None:
    """
    Rename an architecture.
    """
    rename_instance(path, old_name, new_name)

def rename_simulation(path, old_name, new_name) -> None:
    """
    Rename a simulation.
    """
    rename_instance(path, old_name, new_name)

def create_instance(path, name) -> None:
    """
    Create a new architecture or simulation.
    """
    instance_path = os.path.join(path, name)
    os.makedirs(instance_path, exist_ok=True)

def create_architecture(path, name) -> None:
    """
    Create a new architecture.
    """
    create_instance(path, name)

def create_simulation(path, name) -> None:
    """
    Create a new simulation.
    """
    create_instance(path, name)

######################################
# Architecture Settings
######################################
 
def load_architecture_settings(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> dict:
    """
    Load the settings of a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    path = os.path.join(path, hard_settings.param_settings_filename)
    if not os.path.exists(path):
        return {}
    with open(path, "r") as f:
        return yaml.safe_load(f) or {}

def save_architecture_settings(arch_path, arch_name, settings) -> None:
    """
    Save the settings of a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, hard_settings.main_parameter_domain)
    os.makedirs(path, exist_ok=True)
    path = os.path.join(path, hard_settings.param_settings_filename)

    yaml_obj = YAML()

    data = CommentedMap()

    data.yaml_set_start_comment(
f"""##############################################
# Settings for {arch_name}
##############################################

# This file was generated by Odatix GUI {motd.read_version()} on {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}.
# You can still modify manually this file as needed.
"""
    )

    generate_rtl = settings.get('generate_rtl', False)
    data['generate_rtl'] = generate_rtl
    data.yaml_set_comment_before_after_key('generate_rtl', before="\nRTL generation")
    
    if generate_rtl:
        data['design_path'] = settings.get('design_path', [])
        data['design_path_whitelist'] = settings.get('design_path_whitelist', [])
        data['design_path_blacklist'] = settings.get('design_path_blacklist', [])
        data['generate_command'] = settings.get('generate_command', "")
        data['generate_output'] = settings.get('generate_output', "")
    else:
        data['rtl_path'] = settings.get('rtl_path', "")
    data['top_level_file'] = settings.get('top_level_file', "")
    data['top_level_module'] = settings.get('top_level_module', "")
    if generate_rtl:
        data.yaml_set_comment_before_after_key('top_level_file', before="\nSource files")
    else:
        data.yaml_set_comment_before_after_key('rtl_path', before="\nSource files")

    data['clock_signal'] = settings.get('clock_signal', "")
    data['reset_signal'] = settings.get('reset_signal', "")
    data.yaml_set_comment_before_after_key('clock_signal', before="\nSignals")

    data['use_parameters'] = settings.get('use_parameters', False)
    data['start_delimiter'] = settings.get('start_delimiter', "")
    data['stop_delimiter'] = settings.get('stop_delimiter', "")
    data.yaml_set_comment_before_after_key('use_parameters', before="\nDelimiters for parameter files")

    fmax = CommentedMap()
    settings_fmax = settings.get('fmax_synthesis', {})
    lower_bound = settings_fmax.get('lower_bound', "")
    upper_bound = settings_fmax.get('upper_bound', "")
    if lower_bound != "" and upper_bound != "":
        fmax = {}
    elif lower_bound != "":
        fmax['lower_bound'] = lower_bound if lower_bound != "" else hard_settings.default_fmax_lower_bound
    elif upper_bound != "":
        fmax['upper_bound'] = upper_bound if upper_bound != "" else hard_settings.default_fmax_upper_bound
    data['fmax_synthesis'] = fmax
    data.yaml_set_comment_before_after_key('fmax_synthesis', before="\nDefault frequencies (in MHz)")

    custom = CommentedMap()
    settings_custom = settings.get('custom_freq_synthesis', {})
    settings_custom_list = settings_custom.get('list', [])
    custom_list = CommentedSeq(settings_custom_list)
    custom_list.fa.set_flow_style()
    custom['list'] = custom_list
    data['custom_freq_synthesis'] = custom if settings_custom_list else {}

    generate_configurations = settings.get('generate_configurations', False)
    data['generate_configurations'] = generate_configurations
    settings_gen = settings.get('generate_configurations_settings', {})
    if settings_gen:
        data['generate_configurations_settings'] = settings_gen
    data.yaml_set_comment_before_after_key('generate_configurations', before="\nConfiguration generation")

    with open(path, "w") as f:
        yaml_obj.dump(data, f)


######################################
# Parameter Domains
######################################

def get_arch_domain_path(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> str:
    """
    Get the path of a specific parameter domain.
    """
    if domain == hard_settings.main_parameter_domain:
        return os.path.join(arch_path, arch_name)
    else:
        return os.path.join(arch_path, arch_name, domain)

def get_param_domains(arch_path, arch_name) -> list:
    """
    Get the list of parameter domains for a specific architecture.
    """
    folder = os.path.join(arch_path, arch_name)
    if not os.path.isdir(folder):
        return []
    domain_list = [
        d for d in os.listdir(folder)
        if os.path.isdir(os.path.join(folder, d)) and not d.startswith("_")
    ]
    return sorted(domain_list)

def create_parameter_domain(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> None:
    """
    Create a new parameter domain for a specific architecture.
    """
    path = os.path.join(arch_path, arch_name, domain)
    os.makedirs(path, exist_ok=True)

def duplicate_parameter_domain(arch_path, source_arch_name, target_arch_name, source_domain, target_domain) -> None:
    """
    Duplicate a parameter domain for a specific architecture.
    """
    if target_domain == hard_settings.main_parameter_domain:
        raise ValueError("Cannot duplicate to the main parameter domain.")
    if target_domain == source_domain and source_arch_name == target_arch_name:
        raise ValueError("Source and target domain cannot be the same.")
    
    source_path = get_arch_domain_path(arch_path, source_arch_name, source_domain)
    target_path = get_arch_domain_path(arch_path, target_arch_name, target_domain)
    if not os.path.isdir(source_path):
        return
    if os.path.exists(target_path):
        return
    
    if source_domain == hard_settings.main_parameter_domain:
        # copy only the files in the main domain folder, not the folders
        for item in os.listdir(source_path):
            s = os.path.join(source_path, item)
            d = os.path.join(target_path, item)
            if os.path.isdir(s):
                continue
            os.makedirs(target_path, exist_ok=True)
            with open(s, "rb") as fsrc:
                with open(d, "wb") as fdst:
                    fdst.write(fsrc.read())
            # overwrite the settings file to remove unwanted keys
            settings = load_architecture_settings(arch_path, source_arch_name, source_domain)
            save_domain_settings(arch_path, target_arch_name, target_domain, settings)
    else:
        # copy the entire folder
        copytree(source_path, target_path)

def delete_parameter_domain(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> None:
    """
    Delete a specific parameter domain for a specific architecture.
    """
    if domain == hard_settings.main_parameter_domain:
        raise ValueError("Cannot delete the main parameter domain.")
    path = os.path.join(arch_path, arch_name, domain)
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(path)

def rename_parameter_domain(arch_path, arch_name, old_domain, new_domain) -> None:
    """
    Rename a specific parameter domain for a specific architecture.
    """
    if old_domain == hard_settings.main_parameter_domain:
        raise ValueError("Cannot rename the main parameter domain.")
    old_path = os.path.join(arch_path, arch_name, old_domain)
    new_path = os.path.join(arch_path, arch_name, new_domain)
    if not os.path.isdir(old_path):
        return
    if os.path.exists(new_path):
        return
    os.rename(old_path, new_path)

def parameter_domain_exists(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> bool:
    """
    Check if a specific parameter domain exists.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    return os.path.isdir(path)

def check_parameter_domain_use_parameters(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> bool:
    """
    Check if a specific parameter domain uses parameters.
    """
    settings = load_architecture_settings(arch_path, arch_name, domain)
    if not settings:
        return False
    return settings.get('use_parameters', True)

#######################################
# Parameter Domain Settings
#######################################

def save_domain_settings(arch_path, arch_name, domain, settings) -> None:
    """
    Save the settings of a specific parameter domain of an architecture.
    """
    if domain == hard_settings.main_parameter_domain:
        settings = update_raw_settings(arch_path, arch_name, domain, settings)
        save_architecture_settings(arch_path, arch_name, settings)
        return
    
    path = get_arch_domain_path(arch_path, arch_name, domain)
    os.makedirs(path, exist_ok=True)
    path = os.path.join(path, hard_settings.param_settings_filename)

    yaml_obj = YAML()

    data = CommentedMap()

    data.yaml_set_start_comment(
f"""##############################################
# Settings for {arch_name} - {domain}
##############################################

# This file was generated by Odatix GUI {motd.read_version()} on {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}.
# You can still modify manually this file as needed.
"""
    )

    data['use_parameters'] = settings.get('use_parameters', True)
    data['param_target_file'] = settings.get('param_target_file', "")
    data['start_delimiter'] = settings.get('start_delimiter', "")
    data['stop_delimiter'] = settings.get('stop_delimiter', "")
    data.yaml_set_comment_before_after_key('param_target_file', before="\nDelimiters for parameter files")

    data['generate_configurations'] = settings.get('generate_configurations', False)
    settings_gen = settings.get('generate_configurations_settings', {})
    data['generate_configurations_settings'] = settings_gen
    data.yaml_set_comment_before_after_key('generate_configurations', before="\nConfiguration generation settings")

    with open(path, "w") as f:
        yaml_obj.dump(data, f)

def update_domain_settings(arch_path, arch_name, domain, settings_to_update) -> None:
    """
    Update only the provided settings in the YAML file, preserving comments and other keys.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    path = os.path.join(path, hard_settings.param_settings_filename)

    if not os.path.exists(path):
        save_domain_settings(arch_path, arch_name, domain, settings_to_update)
        return

    yaml_obj = YAML()

    # Load current settings with comments
    if os.path.exists(path):
        with open(path, "r") as f:
            settings = yaml_obj.load(f)
            if settings is None:
                settings = CommentedMap()
    else:
        settings = CommentedMap()

    # Update only the provided keys
    for k, v in settings_to_update.items():
        # Special handling for 'generate_configurations_settings' to preserve list formatting
        if k == "generate_configurations_settings" and isinstance(v, dict):
            _compact_list_variables_in_config_settings(v)
        settings[k] = v

    # Save back, preserving comments and formatting
    with open(path, "w") as f:
        yaml_obj.dump(settings, f)


######################################
# Configuration Files
######################################

def get_config_files(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> list:
    """
    Get the list of configuration files for a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    if not os.path.isdir(path):
        return []
    return sorted([
        f for f in os.listdir(path)
        if f.endswith(".txt") and os.path.isfile(os.path.join(path, f))
    ])

def load_config_file(arch_path, arch_name, domain, filename) -> str:
    """
    Load the content of a specific configuration file for a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    path = os.path.join(path, filename)
    if not os.path.exists(path):
        return ""
    with open(path, "r") as f:
        return f.read()

def save_config_file(arch_path, arch_name, domain, filename, content) -> None:
    """
    Save the content of a specific configuration file for a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    os.makedirs(path, exist_ok=True)
    path = os.path.join(path, filename)
    with open(path, "w") as f:
        f.write(content)

def delete_config_file(arch_path, arch_name, domain, filename) -> None:
    """
    Delete a specific configuration file for a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    path = os.path.join(path, filename)
    if os.path.exists(path):
        os.remove(path)

def delete_all_config_files(arch_path, arch_name, domain=hard_settings.main_parameter_domain) -> None:
    """
    Delete all configuration files for a specific parameter domain of an architecture.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    if not os.path.isdir(path):
        return
    for f in os.listdir(path):
        if f.endswith(".txt"):
            os.remove(os.path.join(path, f))

def generate_config_combinations(domains_configs, arch_name) -> list:
    if domains_configs == {}:
        return []

    domain_names = list(domains_configs.keys())
    config_lists = [domains_configs[d] for d in domain_names]

    combinations = []
    for combo in product(*config_lists):
        replaced_parts = []
        for domain, cfg in zip(domain_names, combo):
            display_domain = arch_name if domain == hard_settings.main_parameter_domain else domain
            replaced_parts.append(f"{display_domain}/{cfg}")

        if domain_names and domain_names[0] != hard_settings.main_parameter_domain:
            combination = [arch_name] + replaced_parts
        else:
            combination = replaced_parts

        combinations.append(combination)

    return combinations

#######################################
# Configuration Generation
#######################################

def create_config_gen_dict(name: str, template: str, variables: dict) -> dict:
    """
    Create a configuration generation dictionary.

    Args:
        name (str): Name of the configuration.
        template (str): Template string for the configuration name.
        variables (dict): Dictionary of variables and their values.

    Returns:
        dict: Configuration generation dictionary.
    """
    return {
        'generate_configurations': True,
        'generate_configurations_settings': {
            'name': name,
            'template': template,
            'variables': variables,
        }
    }

def create_config_gen_variable_dict(name: str, type: str, settings: dict, format: Optional[str]=None) -> dict:
    """
    Create a configuration generation variable dictionary.

    Args:
        name (str): Name of the variable.
        type (str): Type of the variable (e.g., 'range', 'list', 'function').
        settings (dict): Settings specific to the variable type.

    Returns:
        dict: Configuration generation variable dictionary.
    """
    var_dict = {
        name: {
            'type': type,
            'settings': settings,
        }
    }
    if format:
        var_dict[name]['format'] = format
    return var_dict


#######################################
# Helpers
#######################################

def _compact_list_variables_in_config_settings(config_settings: dict):
    """
    Edit the 'variables' in 'generate_configurations_settings' to ensure lists are compact in YAML.
    Args:
        config_settings (dict): The configuration settings dictionary.    
    """
    variables = config_settings.get("variables", {})
    for var_name, var_cfg in variables.items():
        if (
            isinstance(var_cfg, dict)
            and var_cfg.get("type") == "list"
            and "settings" in var_cfg
            and isinstance(var_cfg["settings"], dict)
            and "list" in var_cfg["settings"]
        ):
            compact_list = CommentedSeq(var_cfg["settings"]["list"])
            compact_list.fa.set_flow_style()
            var_cfg["settings"]["list"] = compact_list

def update_raw_settings(arch_path, arch_name, domain, settings_to_update) -> dict:
    """
    Update only the provided keys from a yaml file, without any formatting.
    Args:
        arch_path (str): Path to the architectures folder.
        arch_name (str): Name of the architecture.
        domain (str): Name of the parameter domain.
        settings_to_update (dict): Dictionary of settings to update.
    Returns:
        dict: Updated settings.
    """
    path = get_arch_domain_path(arch_path, arch_name, domain)
    path = os.path.join(path, hard_settings.param_settings_filename)

    # Load current settings if the file exists
    if os.path.exists(path):
        with open(path, "r") as f:
            settings = yaml.safe_load(f) or {}
    else:
        settings = {}

    # Update only the provided keys
    for k, v in settings_to_update.items():
        settings[k] = v
    
    return settings



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

"""
This module contains utility functions for file operations, error handling,
and user interactions within Odatix.

Functions:
    - copytree: Custom function to copy directories with filtering options.
    - chunk_list: Splits a list into smaller chunks of a given size.
    - read_from_list: Reads and validates values from a dictionary.
    - move_cursor_up: Moves the cursor up in the terminal.
    - progress_bar: Displays a progress bar in the terminal.
    - ask_to_continue: Prompts the user to continue or exit.
    - ask_yes_no: Prompts the user for a yes/no input.
    - create_dir: Creates a directory, removing it first if it exists.
    - internal_error: Logs internal errors and prints an error message.
    - safe_df_append: Safely appends data to a pandas DataFrame.
    - open_path_in_explorer: Opens a file or directory in the system explorer.
"""

import os
import sys
import glob
import time
import shutil
import fnmatch
import platform
import traceback
import subprocess
from datetime import datetime

import odatix.lib.printc as printc
import odatix.components.motd

YAML_BOOL = ('true', 'false', 'yes', 'no', 'on', 'off')

def copytree(src, dst, dirs_exist_ok=False, whitelist=None, blacklist=None, **kwargs):
  """
  Custom copytree to copy directories with support for dirs_exist_ok, whitelist, and blacklist.

  Args:
    src (str): Source directory.
    dst (str): Destination directory.
    dirs_exist_ok (bool): Whether to allow overwriting the destination directory.
    whitelist (list, optional): List of patterns to include (relative to `src`).
    blacklist (list, optional): List of patterns to exclude (relative to `src`).
    **kwargs: Additional arguments passed to shutil.copy2.
  """
  def is_pattern_matched(path, patterns):
    """
    Check if the path matches any pattern in the list.
    Args:
      path (str): The path to check (relative to `src`).
      patterns (list): List of patterns to match against.
    Returns:
      bool: True if any pattern matches, False otherwise.
    """
    for pattern in patterns:
      if fnmatch.fnmatch(path, pattern):
        return True
      if os.path.realpath(pattern) == os.path.realpath(path):
        return True
      if os.path.dirname(os.path.realpath(path)).startswith(os.path.realpath(pattern)):
        return True
    return False

  def should_copy_file(rel_path):
    """
    Determine if a file should be copied based on whitelist and blacklist.
    Args:
      rel_path (str): Relative path to check.
    Returns:
      bool: True if the file should be copied, False otherwise.
    """
    # If whitelist exists, only copy paths matching the whitelist
    if whitelist and not is_pattern_matched(rel_path, whitelist) and not os.path.isdir(rel_path):
      return False
    # Exclude paths matching the blacklist
    if blacklist and is_pattern_matched(rel_path, blacklist):
      return False
    # Default: copy the path
    return True

  def should_explore_dir(rel_path):
    """
    Determine if a directory should be explored based on the blacklist.
    The whitelist does not apply to directories for exploration.
    Args:
      rel_path (str): Relative path to check.
    Returns:
      bool: True if the directory should be explored, False otherwise.
    """
    # Exclude directories matching the blacklist
    if blacklist and is_pattern_matched(rel_path, blacklist):
      return False
    # Default: explore the directory
    return True

  # Normalize paths
  src = os.path.realpath(src)
  dst = os.path.realpath(dst)

  # Ensure destination exists
  if os.path.exists(dst):
    if not dirs_exist_ok:
      raise FileExistsError(f"Destination directory {dst} exists and dirs_exist_ok is False.")
  else:
    os.makedirs(dst)

  for root, dirs, files in os.walk(src):
    # Calculate the relative path from src
    rel_root = os.path.relpath(root, src)

    # Remove directories that should not be explored
    dirs[:] = [d for d in dirs if should_explore_dir(os.path.join(rel_root, d))]

    # Process files
    for file in files:
      src_file = os.path.join(root, file)
      rel_file = os.path.join(rel_root, file)  # Relative path for whitelist/blacklist checks
      dst_file = os.path.join(dst, rel_file)
      
      if should_copy_file(rel_file):  # Pass relative path to should_copy_file
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        shutil.copy2(src_file, dst_file, **kwargs)

def chunk_list(lst, n):
  """Splits a list into chunks of size n."""
  for i in range(0, len(lst), n):
    yield lst[i:i + n]

class KeyNotInListError(Exception):
  """Exception raised when a required key is missing in a list."""
  pass

class BadValueInListError(Exception):
  """Exception raised when a value in a list is of the wrong type."""
  pass

def read_from_list(key, input_list, filename, raise_if_missing=True, optional=False, print_error=True, parent=None, type=None, script_name=""):
  """
  Retrieves a value from a dictionary, with error handling and type validation.
  """
  parent_string = "" if parent == None else ", inside list \"" + parent + "\","
  if key in input_list:
    value = input_list[key]
    if type is None:
      return value
    else:
      if isinstance(value, type):
        return value
      else:
        if print_error:
          if optional:
            printc.note("Value \"" + str(value) + "\" for key \"" + key + "\"" + parent_string + " in \"" + filename + "\" is of type \"" + value.__class__.__name__ + "\" while it should be of type \"" + type.__name__ + "\". Using default values instead.", script_name)
          else:
            printc.error("Value \"" + str(value) + "\" for key \"" + key + "\"" + parent_string + " in \"" + filename + "\" is of type \"" + value.__class__.__name__ + "\" while it should be of type \"" + type.__name__ + "\".", script_name)
        if raise_if_missing:
          raise BadValueInListError
        return False
  else:
    if print_error:
      if optional:
        printc.note("Cannot find optional key \"" + key + "\"" + parent_string + " in \"" + filename + "\". Using default values instead.", script_name)
      else:
        printc.error("Cannot find key \"" + key + "\"" + parent_string + " in \"" + filename + "\".", script_name)
    if raise_if_missing:
      raise KeyNotInListError
    return False
  
def move_cursor_up():
  """Moves the terminal cursor up one line."""
  sys.stdout.write('\x1b[1A') # Move cursor up
  sys.stdout.write("\033[K") # Clear to the end of line
  sys.stdout.flush()

def progress_bar(progress, title, title_size=50, bar_size=50, endstr='', progress_char="#"):
  """Displays a progress bar in the terminal."""
  if progress is None:
    limit = 1
  else:
    if progress > 100:
      progress = 100
    limit = int(progress * bar_size / 100)

  padded_title = title.ljust(title_size)

  print(padded_title + " [", end='')
  if progress is None:
    printc.color(printc.colors.BLINK)
  for i in range(0, limit):
    print(progress_char, end='')
  for i in range(limit, bar_size):
    print(' ', end='')
  if progress is None:
    printc.color(printc.colors.ENDC)
  print("]", end='')
  if progress is not None:
    print(" {}%".format(int(progress)) + endstr)
  else:
    print()

def ask_to_continue(exit_code=-1):
  print("Continue? ", end="")
  answer = ask_yes_no()
  if answer is False:
    sys.exit(exit_code)
    
def ask_yes_no():
  while True:
    answer = input("(Y/n) ")
    if answer.lower() in ['yes', 'ye', 'y', '1', '']:
      return True
    elif answer.lower() in ['no', 'n', '0']:
      return False
    else:
      print("Please enter yes or no")

def create_dir(dir):
  """Creates a directory, removing it first if it exists."""
  if os.path.isdir(dir):
    for attempt in range(5):  # Try multiple times in case of latency or locked files
      try:
        shutil.rmtree(dir)
        if not os.path.exists(dir):  # Check if deletion was successful
          break
      except (PermissionError, OSError):
        time.sleep(0.5)  # Wait before retrying
    else:
      printc.error(f"Unable to delete {dir} after 5 attempts")

  os.makedirs(dir)

def internal_error(e, error_logfile, script_name):
  tb_full = traceback.format_exc()
  command_line = ' '.join(sys.argv)
  current_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
  system_info = {
    "OS": platform.system(),
    "OS Version": platform.version(),
    "Python Version": platform.python_version(),
    "Machine": platform.machine(),
  }
  with open(error_logfile, "w") as log_file:
    log_file.write("Date and Time: " + current_time + "\n\n")
    log_file.write("System Information:\n")
    for key, value in system_info.items():
      log_file.write("  " + key + ": " + value + "\n")
    log_file.write("\nOdatix Version: " + str(odatix.components.motd.read_version()) + "\n\n")
    log_file.write("\nCommand:\n")
    log_file.write("  " + command_line + "\n\n")
    log_file.write(tb_full)
  printc.internal_error(type(e).__name__ + ": " + str(e), script_name)
  printc.note('Full error details written to "' + error_logfile + '"', script_name)
  printc.note("Please, report this error with the error log attached", script_name)

def safe_df_append(df, row, ignore_index=True):
  if hasattr(df, 'append'):
    return df.append(row, ignore_index=ignore_index)
  else:
    return df._append(row, ignore_index=ignore_index)

def open_path_in_explorer(path):
  """Opens the given path in the system file explorer."""
  if sys.platform.startswith("win"): # Windows
    os.startfile(path)
  elif sys.platform.startswith("linux"): # Linux
    subprocess.run(["xdg-open", path])
  elif sys.platform.startswith("darwin"): # macOS
    subprocess.run(["open", path])
  else:
    raise NotImplementedError(f"Unsupported platform: {platform}")

def merge_dicts_of_lists(base_dict, merging_dict):
  for key, new_list in merging_dict.items():
    if key not in base_dict:
      base_dict[key] = list(new_list)
    else:
      combined = set(base_dict[key]) | set(new_list)
      base_dict[key] = sorted(combined)
  return base_dict

def get_timestamp_string() -> str:
  return datetime.now().strftime("%Y-%m-%d_%H-%M-%S") 



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import html, dcc, Input, Output, State, ctx
from typing import Optional#, Literal

import odatix.components.workspace as workspace
from odatix.gui.icons import icon
from odatix.gui.utils import get_key_from_url
import odatix.gui.ui_components as ui
import odatix.gui.navigation as navigation
import odatix.lib.hard_settings as hard_settings
from odatix.lib.settings import OdatixSettings

page_path = "/arch_editor"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix - Architecture Editor',
    name='Architecture Editor',
    order=3,
)

######################################
# UI Components
######################################

def architecture_title(arch_name):
    variable_title_tile_buttons = html.Div(
        children=[
            ui.icon_button(
                id=f"button-open-config-editor",
                icon=icon("edit", className="icon blue"),
                text="Edit Configs",
                tooltip="Open the Configuration Editor for this architecture",
                tooltip_options="bottom delay",
                color="default",
                link=f"/config_editor?arch={arch_name}",
                multiline=False,
                width="135px",
            ),
            ui.save_button(
                id={"page": page_path, "action": "save-all"},
                tooltip="Save all changes",
                disabled=True,
            ),
        ],
        className="inline-flex-buttons",
    )
    back_btn = ui.back_button(link="/architectures")
    
    return html.Div(
        html.Div(
            children=[
                html.Div(
                    children=[
                        html.Div(
                            children=[
                                dcc.Input(
                                    value=f"{arch_name}",
                                    type="text",
                                    id="arch-title",
                                    placeholder="Architecture Name...",
                                    className="title-input",
                                    style={"width": "100%", "transform": "translate(-5px, 5px)"},
                                )
                            ],
                            id="arch-title-container",
                        ),
                        html.Div(
                            [variable_title_tile_buttons],
                        ),
                    ],
                    className="title-tile-flex",
                    style={
                        "display": "flex",
                        "alignItems": "center",
                        "padding": "0px",
                        "justifyContent": "space-between",
                    }
                ),
                back_btn,
            ],
            className="tile title",
            style={"position": "relative"},
        ),
        className="card-matrix config",
        style={"marginTop": "0px", "marginBottom": "10px", "marginLeft": "-13px"},
    )

def architecture_form_field(
    label: str,
    id: str,
    value: str="",
    tooltip: str="",
    placeholder: str="",
    tooltip_options: str="secondary",
    # type: Optional[Literal["text", "number", "password", "email", "range", "search", "tel", "url", "hidden"]] = None,
    type = None,
):
    return html.Div(
        children=[
            html.Label(label),
            ui.tooltip_icon(tooltip, tooltip_options),
            dcc.Input(id=id, value=value, type=type, placeholder=placeholder, style={"width": "100%"}),
        ],
        style={"marginBottom": "12px"}
    )

def architecture_form(settings):
    defval = lambda k, v=None: settings.get(k, v)
    generate_rtl = True if str(defval("generate_rtl", "No")).lower() in ["yes", "true"] else False
    expand_design_path_filters = True if defval("design_path_whitelist", []) or defval("design_path_blacklist", []) else False

    return html.Div(
        children=[
            html.Div(style={"display": "none"}),
            html.Div([
                html.H3("RTL Generation"),
                html.Div([
                    dcc.Checklist(
                        options=[{"label": "Enable RTL Generation", "value": True}],
                        value=[True] if generate_rtl else [],
                        id="generate_rtl",
                        className="checklist-switch",
                        style={"marginBottom": "12px", "marginTop": "5px", "display": "inline-block"},
                    ),
                    ui.tooltip_icon("You can generate the RTL files from a higher-level design description like Chisel, HLS or any other tool using a custom command."),
                ], style={"marginBottom": "12px"}),
                html.Div(
                    children=[
                        architecture_form_field(
                            label="Design Path",
                            id="design_path",
                            value=defval("design_path", ""),
                            tooltip="Path to the design files needed for RTL generation. The whole path will be copied for each configuration, so use white/blacklists below to filter files if needed.",
                        ),
                        html.Div(
                            children=[
                                architecture_form_field(
                                    label="Design Path Whitelist",
                                    id="design_path_whitelist",
                                    value=", ".join(defval("design_path_whitelist", [])),
                                    placeholder="src, project, build.sbt",
                                    tooltip="Comma-separated list of patterns to include files from the design path. Only files matching these patterns will be included. Leave empty to include all files.",
                                ),
                                architecture_form_field(
                                    label="Design Path Blacklist",
                                    id="design_path_blacklist",
                                    value=", ".join(defval("design_path_blacklist", [])),
                                    placeholder="*.txt, docs/, tmp/",
                                    tooltip="Comma-separated list of patterns to exclude files from the design path. Files matching these patterns will be excluded. Leave empty to exclude no files. Note that you can use both whitelist and blacklist together.",
                                ),
                            ],
                            id="more-fields-design-path-filters",
                            className="animated-section" + ("" if expand_design_path_filters else " hide"),
                        ),
                        architecture_form_field(
                            label="Generation Command",
                            id="generate_command",
                            value=defval("generate_command", ""),
                            tooltip="Command to generate the RTL files. This command will be executed in each copy of the design path directory. Make sure all files needed for this command to succeed are included in the design path.",
                        ),
                        architecture_form_field(
                            label="Generation Output",
                            id="generate_output",
                            value=defval("generate_output", ""),
                            tooltip="Path to the generated RTL files relative to the design path.",
                        ),
                        html.Div(
                            children=[
                                ui.icon_button(
                                    icon=icon(
                                        "more",
                                        className="icon normal rotate" + (" rotated" if expand_design_path_filters else ""),
                                        id="more-fields-design-path-filters-toggle-icon"
                                    ),
                                    color="default",
                                    id="more-fields-design-path-filters-toggle",
                                    tooltip="Show/Hide whitelist and blacklist fields",
                                ),
                            ],
                            style={"marginTop": "20px"},
                        ),
                    ],
                    id="generate-settings",
                    className="animated-section" + ("" if generate_rtl else " hide"),
                    style={"overflow": "visible"},
                ),
            ], className="tile config"),
            html.Div([
                html.H3("Top Level Settings"),
                html.Div(
                    children=[
                        architecture_form_field(
                            label="RTL Path",
                            id="rtl_path",
                            value=defval("rtl_path", ""),
                            tooltip="The path to the directory containing the RTL source files.",
                        )
                    ],
                    id="rtl-path-container",
                    className="animated-section" + ("" if not generate_rtl else " hide"),
                    style={"overflow": "visible"},
                ),
                architecture_form_field(
                    label="Top Level File",
                    id="top_level_file",
                    value=defval("top_level_file", ""),
                    tooltip="The path of the file containing the top level (main) module/entity definition. This path is relative to the RTL path.",
                ),
                architecture_form_field(
                    label="Top Level Module",
                    id="top_level_module",
                    value=defval("top_level_module", ""),
                    tooltip="The name of the top level module/entity in your top level file.",
                ),
                architecture_form_field(
                    label="Clock Signal",
                    id="clock_signal",
                    value=defval("clock_signal", ""),
                    tooltip="The name of the clock signal in your top level module/entity.",
                ),
                architecture_form_field(
                    label="Reset Signal",
                    id="reset_signal",
                    value=defval("reset_signal", ""),
                    tooltip="The name of the reset signal in your top level module/entity.",
                ),
            ], className="tile config"),
            html.Div([
                html.H3("Synthesis Settings"),
                html.H4("Fmax Synthesis (MHz)"),
                architecture_form_field(
                    label="Lower Bound",
                    id="fmax_synthesis_lower",
                    value=defval("fmax_synthesis", {}).get("lower_bound", ""),
                    placeholder=str(hard_settings.default_fmax_lower_bound),
                    tooltip="The lower bound for the synthesis maximum operating frequency binary search ('odatix fmax' command). This value can be overriden by the argument --from (ex: 'odatix fmax --from 50 --to 200').",
                    tooltip_options="secondary large",
                    type="number",
                ),
                architecture_form_field(
                    label="Upper Bound",
                    id="fmax_synthesis_upper",
                    value=defval("fmax_synthesis", {}).get("upper_bound", ""),
                    placeholder=str(hard_settings.default_fmax_upper_bound),
                    tooltip="The upper bound for the synthesis maximum operating frequency binary search ('odatix fmax' command). This value can be overriden by the argument --to (ex: 'odatix fmax --from 50 --to 200')",
                    tooltip_options="secondary large",
                    type="number",
                ),
                html.H4("Custom Freq Synthesis (MHz)"),
                architecture_form_field(
                    label="List",
                    id="custom_freq_synthesis_list",
                    value=", ".join(map(str, defval("custom_freq_synthesis", {}).get("list", []))),
                    placeholder=", ".join(map(str, hard_settings.default_custom_freq_list)),
                    tooltip="Comma-separated list of custom frequencies for synthesis (in MHz). The synthesis will be run for each frequency in this list ('odatix freq' command). Theses values can be overriden by the argument --at (ex: 'odatix freq --at 50 --at 100') or --from, --to and --step (ex: 'odatix freq --from 50 --to 200 --step 10').",
                    tooltip_options="secondary large",
                    type="text",
                ),
            ], className="tile config"),
        ], className="tiles-container config", style={"marginTop": "-10px", "marginBottom": "20px"},
    )


######################################
# Callbacks
######################################

@dash.callback(
    Output("arch-form-container", "children"),
    Output("architecture-initial-settings", "data"),
    Input(f"url_{page_path}", "search"),
    State(f"url_{page_path}", "pathname"),
    State("odatix-settings", "data"),
)
def init_form(search, page, odatix_settings):
    if page != page_path:
        return dash.no_update, dash.no_update

    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        return architecture_form({}), {}

    if arch_name:
        full_settings = workspace.load_architecture_settings(arch_path, arch_name, hard_settings.main_parameter_domain)
        settings ={
            "generate_rtl": full_settings.get("generate_rtl", False),
            "design_path": full_settings.get("design_path", ""),
            "design_path_whitelist": full_settings.get("design_path_whitelist", []),
            "design_path_blacklist": full_settings.get("design_path_blacklist", []),
            "generate_command": full_settings.get("generate_command", ""),
            "generate_output": full_settings.get("generate_output", ""),

            "rtl_path": full_settings.get("rtl_path", ""),
            "top_level_file": full_settings.get("top_level_file", ""),
            "top_level_module": full_settings.get("top_level_module", ""),
            "clock_signal": full_settings.get("clock_signal", ""),
            "reset_signal": full_settings.get("reset_signal", ""),

            "fmax_synthesis": full_settings.get("fmax_synthesis", {}),
            "custom_freq_synthesis": full_settings.get("custom_freq_synthesis", {}),
        }
    else:
        settings = {}
    return architecture_form(settings), settings

@dash.callback(
    Output({"page": page_path, "action": "save-all"}, "className"),
    Output({"page": page_path, "action": "save-all"}, "data-tooltip"),
    Output(f"url_{page_path}", "search"),
    Output("architecture-saved-settings", "data"),
    Input({"page": page_path, "action": "save-all"}, "n_clicks"),
    Input("arch-title", "value"),
    Input("design_path", "value"),
    Input("design_path_whitelist", "value"),
    Input("design_path_blacklist", "value"),
    Input("generate_rtl", "value"),
    Input("generate_command", "value"),
    Input("generate_output", "value"),
    Input("rtl_path", "value"),
    Input("top_level_file", "value"),
    Input("top_level_module", "value"),
    Input("clock_signal", "value"),
    Input("reset_signal", "value"),
    Input("fmax_synthesis_lower", "value"),
    Input("fmax_synthesis_upper", "value"),
    Input("custom_freq_synthesis_list", "value"),
    State(f"url_{page_path}", "search"),
    State(f"url_{page_path}", "pathname"),
    State("architecture-initial-settings", "data"),
    State("architecture-saved-settings", "data"),
    State("odatix-settings", "data"),
    prevent_initial_call=True,
)
def save_and_status(
    n_clicks, arch_title, design_path, whitelist, blacklist, generate_rtl, generate_command, generate_output,
    rtl_path, top_level_file, top_level_module, clock_signal, reset_signal, 
    fmax_lower, fmax_upper, custom_freq_list, search, page, initial_settings, saved_settings,
    odatix_settings,
):
    triggered_id = ctx.triggered_id
    if triggered_id == f"url_{page_path}" and page != page_path:
        return dash.no_update, dash.no_update, dash.no_update

    if saved_settings is None:
        settings = initial_settings
    else:
        settings = saved_settings

    if fmax_lower is None:
        fmax_lower = ""
    if fmax_upper is None:
        fmax_upper = ""
    if custom_freq_list is None:
        custom_freq_list = ""

    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    
    current_settings_subset = {
        "generate_rtl": True if generate_rtl else False,
        "design_path": design_path,
        "design_path_whitelist": [x.strip() for x in whitelist.split(",") if x.strip()],
        "design_path_blacklist": [x.strip() for x in blacklist.split(",") if x.strip()],
        "generate_command": generate_command,
        "generate_output": generate_output,
        "rtl_path": rtl_path,
        "top_level_file": top_level_file,
        "top_level_module": top_level_module,
        "clock_signal": clock_signal,
        "reset_signal": reset_signal,
        "fmax_synthesis": {
            "lower_bound": fmax_lower,
            "upper_bound": fmax_upper,
        } if fmax_lower != "" and fmax_upper != "" else {
            "lower_bound": fmax_lower,
        } if fmax_lower != "" else {
            "upper_bound": fmax_upper,
        } if fmax_upper != "" else {},
        "custom_freq_synthesis": {
            "list": [int(x.strip()) for x in custom_freq_list.split(", ") if x.strip()],
        } if custom_freq_list != "" else {},
    }

    if arch_name:
        current_settings = workspace.update_raw_settings(
            arch_path=arch_path,
            arch_name=arch_name,
            domain=hard_settings.main_parameter_domain,
            settings_to_update=current_settings_subset
        )
    else: 
        current_settings = current_settings_subset

    if not arch_title:
        return "color-button error-status icon-button tooltip bottom", "Architecture name cannot be empty", dash.no_update, saved_settings
    
    for c in hard_settings.invalid_filename_characters:
        if c in arch_title:
            c = "' ' (space)" if c == " " else f"'{c}'"
            return "color-button error-status icon-button tooltip bottom", f"Unauthorized character in architecture name: {c}", dash.no_update, saved_settings

    if triggered_id == {"page": page_path, "action": "save-all"}:
        # Rename architecture if needed
        new_search = dash.no_update
        if arch_title != arch_name:
            old_name = arch_name
            new_name = arch_title
            if workspace.architecture_exists(arch_path, new_name):
                return "color-button error-status icon-button tooltip bottom", f"'{new_name}' already exists", dash.no_update, dash.no_update
            try:
                workspace.rename_architecture(arch_path, old_name, new_name)
                arch_name = new_name
                new_search = f"?arch={new_name}"
            except Exception as e:
                return "color-button error-status icon-button tooltip bottom", "Failed renaming architecture", dash.no_update, dash.no_update
        
        # Create architecture if it does not exist yet
        if not workspace.architecture_exists(arch_path, arch_name):
            workspace.create_architecture(arch_path, arch_name)

        # Save settings
        try:
            workspace.save_architecture_settings(arch_path, arch_name, current_settings)
            return "color-button disabled icon-button tooltip delay bottom small", "Nothing to save", new_search, current_settings
        except Exception as e:
            return "color-button error-status icon-button tooltip bottom small", "Failed to save...", dash.no_update, dash.no_update
    else:
        if current_settings_subset != settings or arch_title != arch_name:
            return "color-button warning icon-button tooltip bottom small tooltip", "Unsaved changes!", dash.no_update, dash.no_update

    return "color-button disabled icon-button tooltip delay bottom small", "Nothing to save", dash.no_update, saved_settings

@dash.callback(
    Output("generate-settings", "className"),
    Output("rtl-path-container", "className"),
    Input("generate_rtl", "value"),
)
def toggle_generate_settings(generate_rtl):
    generate_settings_class = "animated-section" if generate_rtl else "animated-section hide"
    rtl_path_style = "animated-section" if not generate_rtl else "animated-section hide"
    return generate_settings_class, rtl_path_style

@dash.callback(
    Output({"page": page_path, "type": "architecture-title-div"}, "children"),
    Input(f"url_{page_path}", "search"),
)
def update_architecture_title(search):
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        arch_name = ""
    return architecture_title(arch_name)

@dash.callback(
    Output("more-fields-design-path-filters", "className"),
    Output("more-fields-design-path-filters-toggle-icon", "className"),
    Input("more-fields-design-path-filters-toggle", "n_clicks"),
    State("more-fields-design-path-filters", "className"),
    prevent_initial_call=True,
)
def toggle_more_fields(n_clicks, expandable_area_class ):
    if "hide" in expandable_area_class:
        new_expandable_area_class = "animated-section"
        new_icon_class = "icon normal rotate rotated"
    else:
        new_expandable_area_class = "animated-section hide"
        new_icon_class = "icon normal rotate"
    return new_expandable_area_class, new_icon_class


######################################
# Layout
######################################

layout = html.Div(
    children=[
        dcc.Location(id=f"url_{page_path}"),
        html.Div(id={"page": page_path, "type": "architecture-title-div"}, style={"marginTop": "20px"}),
        html.Div(id="arch-form-container"),
        dcc.Store(id="save-state", data=""),
        dcc.Store(id="architecture-initial-settings", data=None),
        dcc.Store(id="architecture-saved-settings", data=None), 
    ],
    className="page-content",
    style={
        "display": "flex",  
        "flexDirection": "column",
        "min-height": f"calc(100vh - {navigation.top_bar_height})",
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import re
import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import plotly.express as px

import odatix.explorer.themes as themes

plot_colors = px.colors.qualitative.Plotly
marker_symbols = ["circle", "square", "diamond", "cross", "x", "triangle-up", "triangle-down", "pentagon", "star"]
marker_symbols_3d = ["circle", "square", "diamond", "cross", "x", "circle-open", "diamond-open", "square-open"]
patterns = ['', '/', 'x', '-', '|', '+', '.', '\\']
greyed_color = "#aaa"

def create_legend_item(label, value, line_style, color, type="arch", marker_symbol=0, draw_line=True, display=True, marker_3d=False):  
  line_color = color if draw_line else "00000000"
  return html.Div(
    [
      dcc.Checklist(
        id=f"checklist-{type}-{label}",
        options=[{"label": "", "value": label}],
        value=value,
        inline=True,
        style={"display": "inline-block", "margin-right": "10px", "text-wrap": "wrap"},
      ),
      html.Div(
        style={
          "display": "inline-block",
          "width": "30px",
          "height": "2px",
          "border-top": f"{line_style} {line_color}",
          "position": "relative",
        },
        children=get_legend_marker_symbol(marker_symbol, color, marker_3d=marker_3d),
      ),
      html.Div(f"{label}", style={"display": "inline-block", "margin-left": "5px"}),
    ],
    id=f"legend-item-{type}-{label}",
    className="legend-item",
    style={"display": "block" if display else "none", "margin-top": "2.5px", "margin-bottom": "2.5px"},
  )

def create_legend_items(explorer, color_mode="Architecture", symbol_mode="Target", marker_3d=False):
  legend_items = [
    create_legend_item(
      label=architecture, 
      value=[architecture],
      line_style="2px dashed",
      color=get_color(i) if color_mode == "Architecture" else "#000",
      type="arch",
      marker_symbol=i if symbol_mode == "Architecture" else 0,
      marker_3d=marker_3d,
    )
    for i, architecture in enumerate(explorer.all_architectures)
  ]
  return legend_items

def create_target_legend_items(explorer, color_mode="Target", symbol_mode="Target", marker_3d=False):
  legend_items = [
    create_legend_item(
      label=target,
      value=[target],
      line_style="2px dashed",
      color=get_color(i) if color_mode == "Target" else "#fff", 
      type="target",
      marker_symbol=i if symbol_mode == "Target" else 0,
      draw_line=False,
      marker_3d=marker_3d,
    )
    for i, target in enumerate(explorer.all_targets)
  ]
  return legend_items

def create_domain_legend_items(explorer, color_mode="Domain", symbol_mode="Domain", marker_3d=False):
  legend_items = []
  for domain in explorer.all_param_domains.keys():
    for i, config in enumerate(explorer.all_param_domains[domain]):
      legend_items.append(
        create_legend_item(
          label=config,
          value=[config],
          line_style="2px dashed",
          color=get_color(i) if color_mode == "Domain" else "#fff", 
          type=f"domains-{domain}",
          marker_symbol=i if symbol_mode == "Domain" else 0,
          draw_line=False,
          marker_3d=marker_3d,
        )
      )
  return legend_items


def setup_callbacks(explorer):

  # Architectures 
  @explorer.app.callback(
    [Output(f"checklist-arch-{architecture}", "value") for architecture in explorer.all_architectures],
    [Input("show-all-architectures", "n_clicks"), Input("hide-all-architectures", "n_clicks")],
    [State(f"checklist-arch-{architecture}", "value") for architecture in explorer.all_architectures],
  )
  def update_arch_checklist_states(show_all_clicks, hide_all_clicks, *current_values):
    ctx = dash.callback_context
    if not ctx.triggered:
      return current_values

    button_id = ctx.triggered[0]["prop_id"].split(".")[0]
    if button_id == "show-all-architectures":
      return [[architecture] for architecture in explorer.all_architectures]
    elif button_id == "hide-all-architectures":
      return [[] for _ in explorer.all_architectures]

    return current_values

  
  # Targets
  @explorer.app.callback(
    [Output(f"checklist-target-{target}", "value") for target in explorer.all_targets],
    [Input("show-all-targets", "n_clicks"), Input("hide-all-targets", "n_clicks")],
    [State(f"checklist-target-{target}", "value") for target in explorer.all_targets],
  )
  def update_target_checklist_states(show_all_clicks, hide_all_clicks, *current_values):
    ctx = dash.callback_context
    if not ctx.triggered:
      return current_values

    button_id = ctx.triggered[0]["prop_id"].split(".")[0]
    if button_id == "show-all-targets":
      return [[target] for target in explorer.all_targets]
    elif button_id == "hide-all-targets":
      return [[] for _ in explorer.all_targets]

    return current_values

  # Architectures 
  @explorer.app.callback(
    Output("custom-legend", "children"),
    Input("url", "pathname"),
    Input("yaml-dropdown", "value"),
    Input("color-mode-dropdown", "value"),
    Input("symbol-mode-dropdown", "value"),
    Input("toggle-unique-architectures", "value"),
    Input("theme-dropdown", "value"),
    [State(f"checklist-arch-{architecture}", "value") for architecture in explorer.all_architectures]
  )
  def update_architecture_legend(pathname, selected_yaml, color_mode, symbol_mode, unique_architectures, theme, *current_values):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return []

    yaml_architectures = explorer.dfs[selected_yaml]["Architecture"].unique()
    legend_items = []

    legend_default_color = themes.get_legend_default_color(theme)

    i_existing = -1 
    for i, (architecture, value) in enumerate(zip(explorer.all_architectures, current_values)):
      if architecture in yaml_architectures:
        display = True
        i_existing += 1
      else:
        display = False

      trace_id = i if unique_architectures else i_existing
      
      color = get_color(trace_id) if color_mode == "architecture" else legend_default_color
      marker_symbol = trace_id if symbol_mode == "architecture" else 0

      legend_item = create_legend_item(
        label=architecture,
        value=value,
        line_style="2px dashed",
        color=color,
        type="arch",
        marker_symbol=marker_symbol,
        draw_line=True,
        display=display,
        marker_3d=pathname in ("/scatter3d"),
      )
      legend_items.append(legend_item)

    return legend_items

  # Targets
  @explorer.app.callback(
    Output(f"target-legend", "children"),
    Input("url", "pathname"),
    Input("yaml-dropdown", "value"),
    Input("color-mode-dropdown", "value"),
    Input("symbol-mode-dropdown", "value"),
    Input("toggle-unique-targets", "value"),
    Input("theme-dropdown", "value"),
    [State(f"checklist-target-{target}", "value") for target in explorer.all_targets],
  )
  def update_target_legend(pathname, selected_yaml, color_mode, symbol_mode, unique_targets, theme, *current_values):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return []

    yaml_targets = explorer.dfs[selected_yaml]["Target"].unique()
    legend_items = []

    legend_default_color = themes.get_legend_default_color(theme)

    i_existing = -1 
    for i, (target, value) in enumerate(zip(explorer.all_targets, current_values)):
      if target in yaml_targets:
        display = True
        i_existing += 1
      else:
        display = False

      trace_id = i if unique_targets else i_existing
      
      color = get_color(trace_id) if color_mode == "target" else legend_default_color
      marker_symbol = trace_id if symbol_mode == "target" else 0

      legend_item = create_legend_item(
        label=target,
        value=value,
        line_style="2px dashed",
        color=color,
        type="target",
        marker_symbol=marker_symbol,
        draw_line=True,
        display=display,
        marker_3d=pathname in ("/scatter3d"),
      )
      legend_items.append(legend_item)

    return legend_items

  # Domains
  @explorer.app.callback(
    Output("domain-legend", "children"),
    Input("url", "pathname"),
    Input("yaml-dropdown", "value"),
    Input("color-mode-dropdown", "value"),
    Input("symbol-mode-dropdown", "value"),
    Input("param-domain-dropdown", "value"),
    Input("dissociate-domain-dropdown", "value"),
    Input("theme-dropdown", "value"),
    [
      State(f"checklist-domains-{domain}-{config}", "value") 
      for domain in explorer.all_param_domains.keys()
      for config in explorer.all_param_domains[domain]
    ],
    )
  def update_domain_legend(pathname, selected_yaml, color_mode, symbol_mode, selected_domain, dissociate_domain, theme, *current_values):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return []

    legend_items = []

    legend_default_color = themes.get_legend_default_color(theme)

    i_existing = -1 
    i_current_value = 0
    for domain in explorer.all_param_domains.keys():
      for i, config in enumerate(explorer.all_param_domains[domain]):
        if domain == selected_domain and domain in explorer.param_domains[selected_yaml] and config in explorer.param_domains[selected_yaml][domain]:
          display = True
          i_existing += 1
        else:
          display = False

        # trace_id = i if unique_domains else i_existing
        trace_id = i
        
        color = get_color(trace_id) if color_mode == "domain_value" and selected_domain == dissociate_domain else legend_default_color
        marker_symbol = trace_id if symbol_mode == "domain_value" and selected_domain == dissociate_domain else 0

        legend_item = create_legend_item(
          label=config,
          value=current_values[i_current_value],
          line_style="2px dashed",
          color=color,
          type=f"domains-{domain}",
          marker_symbol=marker_symbol,
          draw_line=False,
          display=display,
          marker_3d=pathname in ("/scatter3d"),
        )
        legend_items.append(legend_item)
        i_current_value = i_current_value + 1

    return legend_items

  # Parameter domains 
  @explorer.app.callback(
    Output("param-domain-dropdown", "options"),
    Output("dissociate-domain-dropdown", "options"),
    Input("yaml-dropdown", "value"),
  )
  def update_domain_dropdown_options(selected_yaml):
    if len(explorer.all_param_domains) == 0:
      return [], ["None"]
    
    available_values = [{"label": param.replace("__main__", "main"), "value": param} for param in explorer.param_domains[selected_yaml].keys()]
    available_dissociate_values = [{"label": "None", "value": "None"}] + available_values

    return available_values, available_dissociate_values

  @explorer.app.callback(
    Output("param-domain-dropdown", "value"),
    Output("dissociate-domain-dropdown", "value"),
    Output("color-mode-dropdown", "value"),
    Input("param-domain-dropdown", "options"),
    Input("dissociate-domain-dropdown", "options"),
    Input("dissociate-domain-dropdown", "value"),
    State("param-domain-dropdown", "value"),
    State("color-mode-dropdown", "value"),
  )
  def update_domain_dropdown_values(param_options, dissociate_options, dissociate_domain, param_domain, color_mode):
    if type(dissociate_options) is list:
      return dash.no_update, dash.no_update, dash.no_update
    
    ctx = dash.callback_context

    valid_param_domains = [x["value"] for x in param_options]
    valid_dissociate_values = [x["value"] for x in dissociate_options]

    # Synchronize param_domain if dissociate_domain changes
    if ctx.triggered and "dissociate-domain-dropdown" in ctx.triggered[0]["prop_id"]:
      if dissociate_domain in valid_param_domains and dissociate_domain != "None":
        param_domain = dissociate_domain
        color_mode = "domain_value"

    if param_domain not in valid_param_domains:
      param_domain = valid_param_domains[0] if valid_param_domains else "main"
    if dissociate_domain not in valid_dissociate_values:
      dissociate_domain = "None"

    return param_domain, dissociate_domain, color_mode

  @explorer.app.callback(
    [
      Output(f"checklist-domains-{domain}-{config}", "value")
      for domain in explorer.all_param_domains.keys()
      for config in explorer.all_param_domains[domain]
    ],
    [
      Input("show-all-domains", "n_clicks"),
      Input("hide-all-domains", "n_clicks"),
      Input("param-domain-dropdown", "value")
    ],
    [
      State(f"checklist-domains-{domain}-{config}", "value") 
      for domain in explorer.all_param_domains.keys()
      for config in explorer.all_param_domains[domain]
    ],
  )
  def update_domain_checklist_states(show_all_clicks, hide_all_clicks, selected_domain, *current_values):
    if len(explorer.all_param_domains) == 0:
      return

    ctx = dash.callback_context
    if not ctx.triggered:
      return current_values
    
    if selected_domain is None:
      return current_values

    button_id = ctx.triggered[0]["prop_id"].split(".")[0]
    if button_id == "show-all-domains":
      new_values = []
      i_current_value = 0
      for domain in explorer.all_param_domains.keys():
        for config in explorer.all_param_domains[domain]:
          if domain == selected_domain:
            new_values.append([config])
          else:
            new_values.append(current_values[i_current_value])
          i_current_value += 1
      return new_values
    elif button_id == "hide-all-domains":
      new_values = []
      i_current_value = 0
      for domain in explorer.all_param_domains.keys():
        for config in explorer.all_param_domains[domain]:
          if domain == selected_domain:
            new_values.append([])
          else:
            new_values.append(current_values[i_current_value])
          i_current_value += 1
      return new_values

    return current_values

def get_color(i):
  if i == -1:
    return greyed_color
  return plot_colors[i % len(plot_colors)]

def get_marker_symbol(i):
  return marker_symbols[i % len(marker_symbols)]

def get_marker_symbol_3d(i):
  return marker_symbols_3d[i % len(marker_symbols_3d)]

def get_pattern(i):
  return patterns[i % len(patterns)]

def get_legend_marker_symbol(marker_symbol_id, color="white", marker_3d=False):
  marker_symbol_styles = {
    "circle": {"background-color": color, "border-radius": "50%"},
    "square": {"background-color": color, "border-radius": "0"},
    "diamond": {"background-color": color, "transform": "rotate(45deg)", "left": "33%"},
    "cross": {},
    "x": {"left": "40%", "top": "-11px", "font-size": "18px"},
    "cross": {"left": "42%", "top": "-15px"},
    "triangle-up": {"left": "45%", "top": "-19px"},
    "triangle-down": {"left": "45%", "top": "-18px"},
    "pentagon": {"left": "42%", "top": "-10px", "font-size": "15px",},
    "star": {"left": "35%", "top": "-12px", "font-size": "20px",},
    "circle-open": {"border-color": color, "border-style": "solid", "box-sizing": "border-box", "border-radius": "50%"},
    "square-open": {"border-color": color, "border-style": "solid", "box-sizing": "border-box", "border-radius": "0"},
    "diamond-open": {"border-color": color, "border-style": "solid", "box-sizing": "border-box", "transform": "rotate(45deg)", "left": "33%"},
  }

  text = {
    "circle": "",
    "square": "",
    "diamond": "",
    "cross": "+",
    "x": "✖",
    "triangle-up": "▴",
    "triangle-down": "▾",
    "pentagon": "⬟",
    "star": "★",
    "circle-open": "",
    "square-open": "",
    "diamond-open": "",
  }

  try:
    marker_symbol = marker_symbols_3d[marker_symbol_id] if marker_3d else marker_symbols[marker_symbol_id] 
    marker_style = marker_symbol_styles.get(marker_symbol, marker_symbol_styles["circle"])
    text = text.get(marker_symbol, "")
  except Exception as e:
    marker_style = marker_symbol_styles["circle"]
    text = ""

  return html.Div(
    children=html.Div(
      text,
      style={
        "font-weight": "bold",
      }
    ),
    style={
      "font-size": "25px",
      "position": "absolute",
      "top": "-6px",
      "left": "50%",
      "transform": "translateX(-50%)",
      "color": color,
      "width": "10px",
      "height": "10px",
      **marker_style,  # Apply the marker style
    }
  )

def unit_to_html(unit):
  # Regex pattern to match ^(-?\d+) for positive/negative exponents
  pattern = r'\^(-?\d+)'
  html_unit = re.sub(pattern, r'<sup>\1</sup>', unit)

  # Regex pattern to match _(-?\d+) for positive/negative subscripts
  pattern_sub = r'\_(-?\d+)'
  unit_with_sub = re.sub(pattern_sub, r'<sub>\1</sub>', html_unit)

  return unit_with_sub


def clean_configuration_name(config_name, dissociate_domain):
  if dissociate_domain != "none":
    parts = config_name.split("+")
    cleaned_parts = [p for p in parts if not p.startswith(f"{dissociate_domain}_")]
    return "+".join(cleaned_parts)
  return config_name



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import plotly.graph_objs as go

import odatix.explorer.legend as legend
import odatix.explorer.content_lib as content_lib
import odatix.explorer.figures as figures
import odatix.explorer.themes as themes

import odatix.explorer.behaviors.lines as bhv_lines
import odatix.explorer.behaviors.columns as bhv_columns
import odatix.explorer.behaviors.radar as bhv_radar

dim = {
  "default": {
    "page_wide": False,
    "lines": {
      "width_legend": 840,
      "width": 475,
      "height": 500,
    },
    "columns": {
      "width_legend": 840,
      "width": 475,
      "height": 500,
    },
    "radar": {
      "width_legend": 840,
      "width": 475,
      "height": 475,
    }
  },
  "large": {
    "page_wide": False,
    "lines": {
      "width_legend": 1000,
      "width": 635,
      "height": 670,
    },
    "columns": {
      "width_legend": 1000,
      "width": 635,
      "height": 670,
    },
    "radar": {
      "width_legend": 1000,
      "width": 635,
      "height": 635,
    }
  },
  "default_tall": {
    "page_wide": False,
    "lines": {
      "width_legend": 840,
      "width": 475,
      "height": 600,
    },
    "columns": {
      "width_legend": 840,
      "width": 475,
      "height": 600,
    },
    "radar": {
      "width_legend": 840,
      "width": 475,
      "height": 550,
    }
  },
  "large_tall": {
    "page_wide": False,
    "lines": {
      "width_legend": 1000,
      "width": 635,
      "height": 770,
    },
    "columns": {
      "width_legend": 1000,
      "width": 635,
      "height": 770,
    },
    "radar": {
      "width_legend": 1000,
      "width": 635,
      "height": 735,
    }
  },
  "page_wide": {
    "page_wide": True,
    "lines": {
      "height": 770,
    },
    "columns": {
      "height": 770,
    },
    "radar": {
      "height": 735,
    }
  },
}

def setup_callbacks(explorer, all_checklist_inputs, all_architecture_inputs, all_target_inputs, all_domains_inputs):

  @explorer.app.callback(
    Output("radar-graphs", "children"),
    [
      Input("overview-layout-dropdown", "value"),
      Input("yaml-dropdown", "value"),
      Input("results-dropdown", "value"),
      Input("show-all-architectures", "n_clicks"),
      Input("hide-all-architectures", "n_clicks"),
      Input("show-all-targets", "n_clicks"),
      Input("hide-all-targets", "n_clicks"),
      Input("show-all-domains", "n_clicks"),
      Input("hide-all-domains", "n_clicks"),
      Input("toggle-legend", "value"),
      Input("toggle-legendgroup", "value"),
      Input("toggle-title", "value"),
      Input("toggle-lines", "value"),
      Input("toggle-close-line", "value"),
      Input("toggle-connect-gaps", "value"),
      Input("toggle-labels", "value"),
      Input("legend-dropdown", "value"),
      Input("chart-type-dropdown", "value"),
      Input("color-mode-dropdown", "value"),
      Input("symbol-mode-dropdown", "value"),
      Input("dl-format-dropdown", "value"),
      Input("background-dropdown", "value"),
      Input("theme-dropdown", "value"),
      Input("toggle-unique-architectures", "value"),
      Input("toggle-unique-targets", "value"),
      Input("dissociate-domain-dropdown", "value"),
    ]  
    + all_checklist_inputs 
  )
  def update_overview_charts(
    layout,
    selected_yaml,
    selected_results,
    show_all,
    hide_all,
    show_all_targets,
    hide_all_targets,
    show_all_domains,
    hide_all_domains,
    toggle_legend,
    toggle_legendgroup,
    toggle_title,
    toggle_lines,
    toggle_close,
    toggle_connect_gaps,
    toggle_labels,
    legend_dropdown,
    chart_type,
    color_mode,
    symbol_mode,
    dl_format,
    background,
    theme,
    toggle_unique_architectures,
    toggle_unique_targets,
    dissociate_domain,
    *checklist_values,
  ):
    try:
      arch_checklist_values = checklist_values[:len(all_architecture_inputs)]
      target_checklist_values = checklist_values[len(all_architecture_inputs):len(all_architecture_inputs) + len(all_target_inputs)]
      domain_checklist_values = checklist_values[len(all_architecture_inputs) + len(all_target_inputs):]

      if not selected_yaml or selected_yaml not in explorer.dfs:
        return html.Div(className="error", children=[html.Div("Please select a YAML file.")])

      ctx = dash.callback_context
      triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

      if triggered_id in ["show-all-architectures", "hide-all-architectures"]:
        visible_architectures = set(
          explorer.dfs[selected_yaml]["Architecture"].unique() if triggered_id == "show-all-architectures" else []
        )
      else:
        visible_architectures = set(
          architecture for i, architecture in enumerate(explorer.all_architectures) if arch_checklist_values[i] and architecture in explorer.dfs[selected_yaml]["Architecture"].unique()
        )

      if triggered_id in ["show-all-targets", "hide-all-targets"]:
        visible_targets = set(
          explorer.dfs[selected_yaml]["Target"].unique() if triggered_id == "show-all-targets" else []
        )
      else:
        visible_targets = set(
          target for i, target in enumerate(explorer.all_targets) if target_checklist_values[i] and target in explorer.dfs[selected_yaml]["Target"].unique()
        )

      i_current_value = 0
      domains = {}
      for domain in explorer.all_param_domains.keys():
        domains[domain] = {}
        for config in explorer.all_param_domains[domain]:
          visible = True if domain_checklist_values[i_current_value] else False
          domains[domain].update({f'{config}': visible})
          i_current_value += 1

      filtered_df = explorer.dfs[selected_yaml][
        (explorer.dfs[selected_yaml]["Architecture"].isin(visible_architectures))
      ]

      for domain, configs in domains.items():
        if domain in filtered_df.columns:
          filtered_df.loc[:, domain] = filtered_df[domain].fillna("None")
          excluded_values = [config for config, is_visible in configs.items() if not is_visible]

          # Apply exclusion only if there are forbidden values in the column
          if any(filtered_df[domain].isin(excluded_values)):
            filtered_df = filtered_df[~filtered_df[domain].isin(excluded_values)]

      metrics = [col for col in filtered_df.columns if col not in ["Target", "Architecture", "Configuration", "Type", "Frequency"]]

      targets_for_yaml = explorer.dfs[selected_yaml]["Target"].unique()

      yaml_name = os.path.splitext(selected_yaml)[0]

      radar_charts = []

      mode = "lines+markers" if toggle_lines else "markers"

      for metric in metrics:

        if chart_type == "lines":
          fig, metric_display, metric_display_unit = bhv_lines.create_line_graph(
            explorer,
            selected_yaml,
            metric,
            selected_results,
            show_all,
            hide_all,
            show_all_targets,
            hide_all_targets,
            show_all_domains,
            hide_all_domains,
            toggle_legend,
            toggle_legendgroup,
            toggle_title,
            toggle_lines,
            toggle_connect_gaps,
            toggle_labels,
            color_mode,
            symbol_mode,
            dl_format,
            background,
            theme,
            toggle_unique_architectures,
            toggle_unique_targets,
            dissociate_domain,
            arch_checklist_values,
            target_checklist_values,
            domain_checklist_values,
          )
          fig.update_xaxes(
            showticklabels=True if toggle_labels else False
          )
        elif chart_type == "columns":
          fig, metric_display, metric_display_unit = bhv_columns.create_column_graph(
            explorer,
            selected_yaml,
            metric,
            selected_results,
            show_all,
            hide_all,
            show_all_targets,
            hide_all_targets,
            toggle_legend,
            toggle_legendgroup,
            toggle_title,
            toggle_lines,
            toggle_labels,
            color_mode,
            symbol_mode,
            dl_format,
            background,
            theme,
            toggle_unique_architectures,
            toggle_unique_targets,
            dissociate_domain,
            arch_checklist_values,
            target_checklist_values,
            domain_checklist_values,
          )
        elif chart_type == "radar":
          fig, metric_display, metric_display_unit = bhv_radar.create_radar_graph(
            explorer,
            selected_yaml,
            metric,
            selected_results,
            show_all,
            hide_all,
            show_all_targets,
            hide_all_targets,
            show_all_domains,
            hide_all_domains,
            toggle_legend,
            toggle_legendgroup,
            toggle_title,
            toggle_lines,
            toggle_close,
            toggle_connect_gaps,
            toggle_labels,
            color_mode,
            symbol_mode,
            dl_format,
            background,
            theme,
            toggle_unique_architectures,
            toggle_unique_targets,
            dissociate_domain,
            arch_checklist_values,
            target_checklist_values,
            domain_checklist_values,
          )
        else:
          metric_display = ""
          metric_display_unit = ""
          fig = go.Figure()

        page_wide =  dim[layout]["page_wide"]
        if page_wide:
          width = None
        else:
          width_legend = dim[layout][chart_type]["width_legend"]
          width = dim[layout][chart_type]["width"]
          width = width_legend if "show_legend" in legend_dropdown else width
        height = dim[layout][chart_type]["height"]

        fig.update_layout(
          paper_bgcolor=background,
          showlegend="show_legend" in legend_dropdown,
          legend_groupclick="toggleitem",
          margin=dict(l=60, r=60, t=60, b=60),
          title=metric_display if toggle_title else None,
          title_x=0.5,
          width=width,
          height=height,
          template=theme,
          polar_angularaxis_showticklabels=True if toggle_labels else False,
          modebar={
            "bgcolor": themes.get_page_bgcolor(theme, default=None),
            "color": themes.get_button_color(theme, default=None),
            "activecolor": themes.get_button_active_color(theme, default=None),
          }
        )

        filename = "Odatix-{}-{}-{}".format(yaml_name, __name__, metric)
        radar_charts.append(figures.make_figure_div(fig, filename, dl_format, page_wide))

      # Add legend chart
      if "separate_legend" in legend_dropdown:
        fig = go.Figure()
        fig, metric_display, metric_display_unit = bhv_radar.create_radar_graph(
          explorer,
          selected_yaml,
          metric,
          selected_results,
          show_all,
          hide_all,
          show_all_targets,
          hide_all_targets,
          show_all_domains,
          hide_all_domains,
          toggle_legend,
          toggle_legendgroup,
          toggle_title,
          toggle_lines,
          toggle_close,
          toggle_connect_gaps,
          toggle_labels,
          color_mode,
          symbol_mode,
          dl_format,
          background,
          theme,
          toggle_unique_architectures,
          toggle_unique_targets,
          dissociate_domain,
          arch_checklist_values,
          target_checklist_values,
          domain_checklist_values,
          legend_mode=True,
        )
        fig.update_layout(
          polar_bgcolor="rgba(255, 255, 255, 0)",
          paper_bgcolor=background,
          showlegend=True,
          legend_groupclick="toggleitem",
          margin=dict(l=60, r=60, t=60, b=60),
          title="Legend" if toggle_title else "",
          title_x=0.5,
          polar=dict(radialaxis=dict(visible=False), angularaxis=dict(visible=False)),
          autosize=True,
          legend_x=0,
          legend_y=1,
          width=width,
          height=height,
          template=theme,
        )
        radar_charts.append(
          figures.make_figure_div(fig, "Odatix-" + str(__name__) + "-legend", dl_format, remove_zoom=True)
        )

      page_background = themes.get_page_bgcolor(theme)
      return html.Div(radar_charts, style={"display": "flex", "flex-wrap": "wrap", "justify-content": "space-evenly", "backgroundColor": page_background})
    except Exception as e:
      return content_lib.generate_error_div(e)
      


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re

default_theme = "odatix"
_regex_theme = r"\.theme\.(\S+)\s*{"

# Fetch themes from the CSS file
_current_dir = os.path.dirname(os.path.abspath(__file__))
_css_file = os.path.join(_current_dir, "assets", "themes.css")

with open(_css_file, "r") as f:
    _content = f.read()
    _available_theme = re.findall(_regex_theme, _content)

_available_theme = list(_available_theme)
list = _available_theme
